%*-------------------------------------------------------------------------------------*
%|                                                                                     |
%|  Except where otherwise noted, this work is licensed under a                        |
%|  Creative Commons Attribution-NonCommercial-ShareAlike 4.0                          |
%|  International (CC BY-NC-SA 4.0) license.                                           |
%|                                                                                     |
%|  Copyright (c) 2018 by Paul Scherrer Institute (http://www.psi.ch)                  |
%|                                                                                     |
%|      Author: CXS group, PSI                                                         |
%*------------------------------------------------------------------------------------*
% You may use this code with the following provisions:
%
% If this code, or subfunctions or parts of it, is used for research in a
%   publication or if it is fully or partially rewritten for another
%   computing language the authors and institution should be acknowledged
%   in written form and additionally you should cite:
%     M. Liebi, M. Georgiadis, A. Menzel, P. Schneider, J. Kohlbrecher,
%     O. Bunk, and M. Guizar-Sicairos, “Nanostructure surveys of
%     macroscopic specimens by small-angle scattering tensor tomography,”
%     Nature 527, 349-352 (2015).   (doi:10.1038/nature16056)
% and
%     M. Liebi, M. Georgiadis, J. Kohlbrecher, M. Holler, J. Raabe, I. 
%     Usov, A. Menzel, P. Schneider, O. Bunk and M. Guizar-Sicairos,
%     "Small-angle X-ray scattering tensor tomography: model of the 
%     three-dimensional reciprocal-space map, reconstruction algorithm 
%     and angular sampling requirements," Acta Cryst. A74, 12-24 (2018). 
%     (doi:10.1107/S205327331701614X)
%
% A publication that focuses on describing features, or parameters, that
%    are already existing in the code should be first discussed with the
%    authors.
%   
% This code and subroutines are part of a continuous development, they
%    are provided “as they are” without guarantees or liability on part
%    of PSI or the authors. It is the user responsibility to ensure its
%    proper use and the correctness of the results.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [params, theta_out1, phi_out1, a_out1] = ...
    optimize_SH(projection, params, theta_0_init, phi_0_init, a_init)

% set default parameters in case they are missing
if ~isfield(params,'avoid_wrapping')
    params.avoid_wrapping = 1;
    fprintf('params.avoid_wrapping is empty, setting to default value params.avoid_wrapping = %d \n', params.avoid_wrapping)
end
if ~isfield(params,'a1_const')
    params.a1_const = 0;
    fprintf('params.a1_const is empty, setting to default value params.a1_const = %d \n', params.a1_const)
end
if ~isfield(params,'regularization')
    params.regularization = 0;
    fprintf('params.regularization is empty, setting to default value params.regularization = %d \n', params.regularization)
end
if ~isfield(params,'mask3D')
    params.mask3D = ones(params.points_per_line,params.lines,params.lines);
    fprintf('params.mask3D is empty, setting to all ones \n')
end
if ~isfield(params,'regularization_angle')
    params.regularization_angle = 0;
    fprintf('params.regularization_angle is empty, setting to default value params.regularization_angle = %d \n', params.regularization_angle)
end
if ~isfield(params,'coeff_soft_limits_low')
    params.coeff_soft_limits_low = -Inf * ones(size(params.l));
    fprintf('params.coeff_soft_limits_low is empty, setting all to default value of -Inf \n');
else
    if numel(params.coeff_soft_limits_low) == 1
        params.coeff_soft_limits_low = params.coeff_soft_limits_low * ones(size(params.l));
    end
end
if ~isfield(params,'coeff_soft_limits_high')
    params.coeff_soft_limits_high = Inf * ones(size(params.l));
    fprintf('params.coeff_soft_limits_high is empty, setting all to default value of Inf \n')
else
    if numel(params.coeff_soft_limits_high) == 1
        params.coeff_soft_limits_high = params.coeff_soft_limits_high * ones(size(params.l));
    end
end
if ~isfield(params,'soft_limit_weight_coeff')
    params.soft_limit_weight_coeff = 1e5 * ones(size(params.l));
    fprintf('params.soft_limit_weight_coeff is empty, setting to default value of 1e5 \n')
else
    if numel(params.soft_limit_weight_coeff) == 1
        params.soft_limit_weight_coeff = params.soft_limit_weight_coeff * ones(size(params.l));
    end
end
if ~isfield(params,'display')
    params.display = 1;
    fprintf('params.display is empty, setting to default value params.display = %d \n', params.display)
end
if ~isfield(params,'dispinter')
    params.dispinter = 1;
    fprintf('params.dispinter is empty, setting to default value params.dispinter = %d \n', params.dispinter)
end
if ~isfield(params,'plot')
    params.plot = 1;
    fprintf('params.plot is empty, setting to default value params.plot = %d \n', params.plot)
end
if ~isfield(params,'volume_upsampling')
    params.volume_upsampling = 1;
    fprintf('params.volume_upsampling is empty, setting to default value params.volume_upsampling = %d \n', params.volume_upsampling)
end
if ~isfield(params,'method')
    params.method = 'bilinear';
    fprintf('params.method is empty, setting to default value params.method = %s \n', params.method)
end
if ~isfield(params,'filter_2D')
    params.filter_2D = 1;
    fprintf('params.filter_2D is empty, setting to default value params.filter_2D = %d \n', params.filter_2D)
end
if ~isfield(params,'slice')
    params.slice = 0;
    fprintf('params.slice is empty, setting to default value params.slice = %d \n', params.slice)
end

numOfvoxels = params.numOfvoxels;

% Arranging the input for optimization
opt_inputs = []; % Vector with parameters to optimize
if params.find_orientation
    opt_inputs = [theta_0_init phi_0_init];
else
    params.theta_0 = theta_0_init;
    params.phi_0 = phi_0_init;
end
if params.find_coefficients
    % in this case only the a_init
    opt_inputs = [opt_inputs a_init];
else
    params.a = a_init;
end

opt_projection = projection(1:params.skip_projections:length(projection));

% optimization
tic
errorplot; %clears the persistent variable for error storage
%%%%%optimization with gradient
itmax = params.itmax; %maximum number of iteration (empty for default = 50)
ftol = [];  %relative function tolerance (empty for default = 1e-3)
xtol = [];  %absolute solution tolerance (empty for default = 1e-3)
opt_out1 = cgmin1('SAXS_tomo_3D_err_metric', opt_inputs, itmax, ftol, xtol, params, opt_projection);
toc

% Rearranging the solution
if params.find_orientation
    theta_out1 = opt_out1(1:numOfvoxels);
    phi_out1 = opt_out1((numOfvoxels+1):(2*numOfvoxels));
    opt_out1 = opt_out1((2*numOfvoxels+1:end));
else
    theta_out1 = params.theta_0;
    phi_out1 = params.phi_0;
end

if params.find_coefficients
    a_out1 = opt_out1;
    if params.a1_const
        a_out1 = [params.a(1:numOfvoxels) a_out1]; % If first coefficient is constant we need to preppend it to the optimized coeffients before output
    end
else
    a_out1 = params.a;
end
