function [E, grad] = SAXS_tomo_3D_err_metric(opt_inputs, params, projection)
%*-------------------------------------------------------------------------------------*
%|                                                                                                           |
%|  Except where otherwise noted, this work is licensed under a            |
%|  Creative Commons Attribution-NonCommercial-ShareAlike 4.0        |
%|  International (CC BY-NC-SA 4.0) license.                                         |
%|                                                                                                           |
%|  Copyright (c) 2017 by Paul Scherrer Institute (http://www.psi.ch)       |
%|                                                                                                           |
%|      Author: CXS group, PSI                                                                |
%*------------------------------------------------------------------------------------*
% You may use this code with the following provisions:
%
% If this code, or subfunctions or parts of it, is used for research in a 
%   publication or if it is fully or partially rewritten for another 
%   computing language the authors and institution should be acknowledged 
%   in written form and additionally you should cite:
%     M. Liebi, M. Georgiadis, A. Menzel, P. Schneider, J. Kohlbrecher, 
%     O. Bunk, and M. Guizar-Sicairos, “Nanostructure surveys of 
%     macroscopic specimens by small-angle scattering tensor tomography,”
%     Nature 527, 349-352 (2015).   (doi:10.1038/nature16056)
%
% A publication that focuses on describing features, or parameters, that
%    are already existing in the code should be first discussed with the
%    authors.
%   
% This code and subroutines are part of a continuous development, they 
%    are provided “as they are” without guarantees or liability on part
%    of PSI or the authors. It is the user responsibility to ensure its 
%    proper use and the correctness of the results.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if numel(params.l) > 4
    error('add more coefficients in error metric');
end

% start a parallel pool (if none exists yet) with a number of workers equal to the number
% of cores on the current machine 
if isempty(gcp('nocreate'))
    parpool('local', feature('numcores'));
end

% parameters for the optimization
find_coefficients = params.find_coefficients;
find_orientation = params.find_orientation;
find_grad = (nargout == 2);

phi_det = params.phi_det;                           % Measurement angles (detector)
l = params.l;                                       % Polar order of the spherical harmonics
m = params.m;
order = numel(l);
lines = params.lines;                            % dimension of scanned area
points_per_line = params.points_per_line;        % dimension of scanned area
numOfsegments = params.numsegments;                      % number of segements the saxs data is averaged (default is 16)
numOfvoxels = params.numOfvoxels;            % dimension of tomogram
mask3D = params.mask3D;
numOfpixels = length(projection)*points_per_line*lines;

%parameters for arb_projection
p.volume_upsampling = params.volume_upsampling;     % Upsamples the volume by a factor of 2 to reduce artifacts (recommended setting =1)
p.method = params.method;                           % 'nearest' or 'bilinear' method for volume allocation
p.filter_2D = params.filter_2D;                     % Strength of filter applied to image (recommended setting =1)

%opt_inputs is the optimization vector, containing theta_struct, phi_struct and a, the coefficents of the spherical harmonics
if find_orientation
    theta_struct = opt_inputs(1:numOfvoxels);
    phi_struct = opt_inputs((numOfvoxels+1):(2*numOfvoxels));
    opt_inputs = opt_inputs((2*numOfvoxels+1):end);
    
    %avoid wrapping of angles
    if params.avoid_wrapping
        phi_struct = mod(phi_struct, 2*pi);
        theta_struct = mod(theta_struct, 2*pi);
    end
else
    theta_struct = params.theta_0;
    phi_struct = params.phi_0;
end

if find_coefficients
    % in case there is a constriction for the values of a (coefficients)
    if params.a1_const
        a = zeros(numOfvoxels, order);
        a(:, 1) = (params.a1)';
        a(:, 2:order) = reshape(opt_inputs, numOfvoxels, order-1);
    else
        a = reshape(opt_inputs, numOfvoxels, order);
    end
    
else
    a = reshape(params.a, numOfvoxels, order);
end

%initalize arrays for the gradients
grad_a = zeros(points_per_line, lines, lines, order);

grad_theta_struct = zeros(points_per_line, lines, lines);
grad_phi_struct = zeros(points_per_line, lines, lines);

sin_theta_struct = reshape(sin(theta_struct), 1, 1, numOfvoxels);
cos_theta_struct = reshape(cos(theta_struct), 1, 1, numOfvoxels);
sin_phi_struct = reshape(sin(phi_struct), 1, 1, numOfvoxels);
cos_phi_struct = reshape(cos(phi_struct), 1, 1, numOfvoxels);

zeros_struct = zeros(1, 1, numOfvoxels);

%%% Define the rotation matrices
% Rot_str is the combination of 1. Rotation around y with theta_struct
% and 2. Rotation around z with theta_struct
% this rotation is needed from the spherical harmonics coordinate system (main
% orientation axis along z-axis) to the object coordinate system
Rot_str = [ ...
    cos_theta_struct.*cos_phi_struct, cos_theta_struct.*sin_phi_struct, -sin_theta_struct; ...
    -sin_phi_struct                 , cos_phi_struct                  , zeros_struct     ; ...
    sin_theta_struct.*cos_phi_struct, sin_theta_struct.*sin_phi_struct, cos_theta_struct];

% the derivative of the rotation matrix with respect to theta_struct and phi_struct
Rot_str_diff_theta = [ ...
    -sin_theta_struct.*cos_phi_struct, -sin_theta_struct.*sin_phi_struct, -cos_theta_struct ; ...
    zeros_struct                     , zeros_struct                     , zeros_struct      ; ...
    cos_theta_struct.*cos_phi_struct , cos_theta_struct.*sin_phi_struct , -sin_theta_struct];

Rot_str_diff_phi = [ ...
    -cos_theta_struct.*sin_phi_struct, cos_theta_struct.*cos_phi_struct, zeros_struct ; ...
    -cos_phi_struct                  , -sin_phi_struct                 , zeros_struct ; ...
    -sin_theta_struct.*sin_phi_struct, sin_theta_struct.*cos_phi_struct, zeros_struct];

% Determine Ylm and Ylm+1 coefficients
Ylm_coef = ...
    [sqrt(1/pi)*[  1/2,      0,       0,      0]; ...
    sqrt(5/pi) *[ -1/4,    3/4,       0,      0]; ...
    sqrt(1/pi) *[ 9/16,  -45/8,  105/16,      0]; ...
    sqrt(13/pi)*[-5/32, 105/32, -315/32, 231/32]];
Ylm_coef = Ylm_coef(1:order, 1:order);

YlmPLUS1_coef = ...
    [                [0,     0,     0,      0]; ...
    sqrt(15/(2*pi)) *[0,  -1/2,     0,      0]; ...
    sqrt(5/pi)      *[0,   9/8, -21/8,      0]; ...
    sqrt(273/(2*pi))*[0, -5/16,  15/8, -33/16]];
YlmPLUS1_coef = YlmPLUS1_coef(1:order, 1:order);

a_temp = reshape(a.', 1, order, []);

a_temp1 = a .* m * Ylm_coef;
a_temp1 = reshape(a_temp1.', 1, order, []);

a_temp2 = a .* sqrt((l-m).*(l+m+1)) * YlmPLUS1_coef;
a_temp2 = reshape(a_temp2.', 1, order, []);

ones_struct = ones(1, numOfsegments, numOfvoxels);

% optimization
E = 0;

% Precalculate parameters
arr_phi_det = [cos(phi_det); sin(phi_det)];

%calculate for all projections
parfor ii = 1:length(projection) %use parallel processing for the loop over all projections
    data = double(projection(ii).data);
    Rot_exp_now = double(projection(ii).Rot_exp);
    Rot_exp_now_mod = Rot_exp_now(1:2,: )' * arr_phi_det; % to be able to multiply
    
    sh_cut = bsxpagemult(double(Rot_str), double(Rot_exp_now_mod));
    cos_theta_sh_cut = sh_cut(3, :, :);
    
    block = double([ones_struct; cumprod(repmat(cos_theta_sh_cut.^2, order-1, 1), 1)]);
    
    Ylm = bsxpagemult(double(Ylm_coef), block);
    Ymn_2D = bsxpagemult(double(a_temp), Ylm);
    
    data_synt_vol = permute(abs(Ymn_2D.^2), [3, 2, 1]); % modeled intensities
    data_synt_vol = reshape(data_synt_vol, points_per_line, lines, lines, numOfsegments);
    
    %%%%projection of the data_synt_vol onto detector plane, in direction of
    %Rot_exp
    N = [points_per_line, lines, lines]; % Size of tomogram
    x = (1:N(2)) - ceil(N(2)/2);
    y = (1:N(1)) - ceil(N(1)/2);
    z = (1:N(3)) - ceil(N(3)/2);
    [X, Y, Z] = meshgrid(x, y, z); % (x,y,z) (2,1,3)
    % If x is defined as the 3rd dimension N(3), then it should appear in the
    % third position of the meshgrid command. If x is defined as the first
    % dimension N(1), then it appears in the second position of the meshgrid,
    % because meshgrid is (x,y,z) and size is (y,x,z)
    
    % output of projection should have the size of the measured projection
    %put in alignment correction values from registration
    xout = (1:size(data,2)) - ceil(size(data,2)/2) + projection(ii).dx;
    yout = (1:size(data,1)) - ceil(size(data,1)/2) + projection(ii).dy;
    
    [proj_out_all, xout, yout] = arb_projection(data_synt_vol, X, Y, Z, Rot_exp_now, p, xout, yout);
    data_synt_vol = []; % free up memory
    
    %%%caluculating the error (difference between estimated intensity and measured data), taking poisson noise into account
    %auxillary functions, used in E and the gradient
    aux_diff_poisson = (sqrt(proj_out_all) - sqrt(data)) .* projection(ii).window_mask;
    E = E + 2*sum(sum(sum(aux_diff_poisson.^2))) / numOfpixels; % error metric with poisson noise
    
    %%%the gradients
    if find_grad
        aux_grad_poisson =  aux_diff_poisson ./ (sqrt(proj_out_all) + 1e-10) / numOfpixels; % maybe add small number to avoid divison by zero
        aux_grad_poisson_vol = arb_back_projection(aux_grad_poisson, xout, yout, X, Y, Z, Rot_exp_now, p);
        
        if find_coefficients
            Ymn_aux_vol = permute(real(repmat(conj(Ymn_2D), order, 1, 1) .* Ylm), [3, 2, 1]);
            Ymn_aux_vol = reshape(Ymn_aux_vol, points_per_line, lines, lines, numOfsegments, order);
            grad_a = grad_a + 4*squeeze(sum(aux_grad_poisson_vol.*Ymn_aux_vol, 4));
            Ymn_aux_vol = []; % free up memory
        end
        Ylm = []; %#ok<*NASGU> % free up memory
        
        if find_orientation
            Rot_diff_theta = bsxpagemult(Rot_str_diff_theta, Rot_exp_now_mod);
            Rot_diff_phi = bsxpagemult(Rot_str_diff_phi, Rot_exp_now_mod);
            
            x_sh_cut = sh_cut(1, :, :);
            y_sh_cut = sh_cut(2, :, :);
            sh_cut = []; % free up memory
            
            sin_theta_sh_cut = sqrt(1-cos_theta_sh_cut.^2);
            phi_sh_cut = atan2(y_sh_cut, x_sh_cut);
            
            const_val = exp(1i.*phi_sh_cut) .* sin_theta_sh_cut .* cos_theta_sh_cut;
            block2 = double([ones_struct; repmat(const_val, order-1, 1, 1) .* block(1:order-1, :, :)]);
            
            const_val = [];
            YlmPLUS1 = bsxpagemult(double(a_temp2), block2);
            block2 = []; % free up memory
            
            if any(m) % full calculations
                Ylm_temp = bsxpagemult(a_temp1, block);
                
                %auxillary variables for gradient of error matrix in respect to
                %theta_struct and phi_struct in each voxel
                temp_theta_1 = Rot_diff_theta(2, :, :) .* x_sh_cut - Rot_diff_theta(1, :, :) .* y_sh_cut ./ x_sh_cut.^2;
                temp_theta_2 = Rot_diff_theta(3, :, :);
                
                temp_phi_1 = Rot_diff_phi(2, :, :) .* x_sh_cut - Rot_diff_phi(1, :, :) .* y_sh_cut ./ x_sh_cut.^2;
                temp_phi_2 = Rot_diff_phi(3, :, :);
                
                expr1 = Ylm_temp .* (cos(phi_sh_cut).^2)*1i;
                expr2 = Ylm_temp .* (cos_theta_sh_cut./sin_theta_sh_cut.^2) + ...
                    YlmPLUS1 .* (exp(-1i*phi_sh_cut)./sin_theta_sh_cut);
                
                YlmPLUS1 = []; % free up memory
                phi_sh_cut = [];
                sin_theta_sh_cut = [];
                block = [];
                
                %auxillary variable for the gradient (numOfvoxels x numofsegments)
                grad_theta_struct_aux_vol = real((expr1 .* temp_theta_1 - expr2 .* temp_theta_2) .* conj(Ymn_2D));
                
                %auxillary variable for the gradient (numOfvoxels x numofsegments)
                grad_phi_struct_aux_vol = real((expr1 .* temp_phi_1 - expr2 .* temp_phi_2) .* conj(Ymn_2D));
                
                expr1 = [];
                expr2 = [];
                
            else % optimized version
                %auxillary variables for gradient of error matrix in respect to
                %theta_struct and phi_struct in each voxel
                expr2 = - YlmPLUS1 .* exp(-1i*phi_sh_cut) ./ sin_theta_sh_cut .* conj(Ymn_2D);
                
                YlmPLUS1 = []; % free up memory
                phi_sh_cut = [];
                sin_theta_sh_cut = [];
                block = [];
                
                %auxillary variable for the gradient (numOfvoxels x numofsegments)
                grad_theta_struct_aux_vol = real(expr2 .* Rot_diff_theta(3, :, :));
                
                %auxillary variable for the gradient (numOfvoxels x numofsegments)
                grad_phi_struct_aux_vol = real(expr2 .* Rot_diff_phi(3, :, :));
                
                expr2 = [];
            end
            
            Rot_diff_theta = [];
            Rot_diff_phi = [];
            
            grad_theta_struct_aux_vol = permute(grad_theta_struct_aux_vol, [3, 2, 1]);
            grad_phi_struct_aux_vol = permute(grad_phi_struct_aux_vol, [3, 2, 1]);
            
            grad_theta_struct_aux_vol = reshape(grad_theta_struct_aux_vol,points_per_line, lines, lines, numOfsegments);
            grad_phi_struct_aux_vol = reshape(grad_phi_struct_aux_vol, points_per_line, lines, lines, numOfsegments);
            
            %%%%%gradient of error matrix in respect to theta_struct and phi_struct, with
            %%%  error metric e=2*(sqrt(I(estimate))-sqrt(I(data)))^2) poisson noise
            %%%  and data_synt=abs((Ymn_2D).^2), for general m
            grad_theta_struct = grad_theta_struct + 4*sum(aux_grad_poisson_vol.*grad_theta_struct_aux_vol, 4);
            grad_phi_struct = grad_phi_struct + 4*sum(aux_grad_poisson_vol.*grad_phi_struct_aux_vol, 4);
            
            aux_grad_poisson_vol = []; % free up memory
            grad_theta_struct_aux_vol = [];
            grad_phi_struct_aux_vol = [];
        end
    end
    
    Ymn_2D = []; % free up memory
end

E_temp = E;
if find_grad
    e = errorplot(E);
    iteration = length(e);
    fprintf('*************************************************** \nIteration %d \n', iteration);
else
    fprintf('Line Search \n')
end

if find_orientation && params.regularization_angle
    if find_grad
        fprintf('The regularization parameter is %.0f \n', params.regularization_angle_coeff)
        [Reg, grad_theta_reg, grad_phi_reg] = angle_regularization(...
            theta_struct, phi_struct, sin_theta_struct, cos_theta_struct, sin_phi_struct, cos_phi_struct, ...
            points_per_line, lines);
        grad_theta_struct = grad_theta_struct + params.regularization_angle_coeff.*grad_theta_reg./numOfvoxels;
        grad_phi_struct = grad_phi_struct + params.regularization_angle_coeff.*grad_phi_reg./numOfvoxels;
    else
        Reg = angle_regularization(...
            theta_struct, phi_struct, sin_theta_struct, cos_theta_struct, sin_phi_struct, cos_phi_struct, ...
            points_per_line, lines);
    end
    
    reg_error_temp = (params.regularization_angle_coeff.*Reg)/numOfvoxels;
    E = E + reg_error_temp;
    fprintf('          total_error = data_error + regul_error \n');
    fprintf('          %f   = %f  + %f \n', E, E_temp, reg_error_temp);
else
    fprintf('          data_error = %f \n', E_temp);
end

if find_grad
    %sieves regularization, blurring the gradient
    if find_coefficients && params.regularization
        if params.slice
            grad_a(:,:,params.slice_pos,:) = convn(grad_a(:,:,params.slice_pos,:), params.kernel, 'same');
        else
            grad_a = convn(grad_a, params.kernel, 'same');
        end
    end
    
    % mask gradients
    if find_coefficients
        grad_a = grad_a .* mask3D;
    end
    
    if find_orientation
        grad_theta_struct = grad_theta_struct .* mask3D;
        grad_phi_struct = grad_phi_struct .* mask3D;
    end
end

if find_coefficients
    a_tomo = reshape(a, points_per_line, lines, lines, order);
end

if find_grad
    % plotting
    if (params.display ~= 0) && (mod(iteration, params.dispinter) == 0)
        if params.plot
            figure(102)
            hold on
            subplot(2, 2, 1)
            
            if find_coefficients
                imagesc(sum(grad_a(:,:,:,1),3)); %plot the first coefficent
                title(sprintf('gradient of a0,2 \n data_error = %0.2d ', E_temp),  'interpreter','none')
            else
                imagesc(real(sum(grad_theta_struct,3))); %plot the first coefficent
                title(sprintf('Polar angle gradient \n data_error = %0.3f ',  E_temp),  'interpreter','none')
            end
            
            axis xy equal tight
            colorbar
            colormap jet
            drawnow
            subplot(2, 2, 2)
            
            if find_coefficients
                if order > 1
                    t =(sum(((a_tomo.*spherical_harm(0,0,0,0)).^2),4)); %plot the first coefficent
                    imagesc(sum(t, 3))
                else
                    imagesc(sum(((a_tomo.*spherical_harm(0,0,0,0)).^2),3)); %plot the first coefficent
                end
                title(sprintf('Calculated SH \n data_error = %0.2d ',   E_temp),  'interpreter','none')
            else
                imagesc(real(sum(grad_phi_struct,3))); %plot the first coefficent
                title(sprintf('Azimuthal angle gradient \n data_error = %0.3f ',   E_temp), 'interpreter','none')
            end
            
            axis xy equal tight
            colorbar
            colormap jet
            drawnow
            subplot(2, 2,[3, 4])
            plot(iteration, E_temp, 'o--b');
            
            if find_orientation
                if params.regularization_angle
                    hold on
                    loglog(iteration, reg_error_temp, 'o--r');
                    hold off
                    legend({'data_error', 'regul_error'},'interpreter','none');
                else
                    legend({'data_error'},'interpreter','none');
                end
                
                if find_coefficients
                    title('Angles and coefficients optimization')
                else
                    title('Angles optimization')
                end
            else
                title('Coefficient optimization')
            end
            
            xlabel('Iteration number');
            ylabel('Error');
            grid on
            
        end
    end
end

% Apply soft limits to the coefficients
if find_coefficients
    for i = 1:order
        a_slice = a_tomo(:,:,:,i);

        ind_low = a_slice < params.coeff_soft_limits_low(i);
        ind_high = a_slice > params.coeff_soft_limits_high(i);

        a_slice(ind_low) = a_slice(ind_low) - params.coeff_soft_limits_low(i);
        a_slice(ind_high) = a_slice(ind_high) - params.coeff_soft_limits_high(i);
        a_slice(~ind_low & ~ind_high) = 0;

        E = E + params.soft_limit_weight_coeff(i)*sum(a_slice(:).^2);

        if find_grad
            grad_a(:,:,:,i) = grad_a(:,:,:,i) + 2*params.soft_limit_weight_coeff(i)*a_slice;
        end
    end
end

if find_grad
    % gradient output vector
    grad = [];
    if find_orientation
        grad_theta_struct = reshape(grad_theta_struct, 1, []);
        grad_phi_struct = reshape(grad_phi_struct, 1, []);
        grad = [grad, grad_theta_struct, grad_phi_struct];
    end

    if find_coefficients
        if params.a1_const
            grad_a(:,:,:,1) = [];
        end
        
        grad_a = reshape(grad_a, 1, []);
        grad = [grad, grad_a];
    end

    % saving intermediate results
    filepath = params.optimization_output_path;
    if find_orientation && find_coefficients
        save([filepath 'current_opt_saxs_all.mat'], 'a_tomo', 'grad_a', ...
            'grad_theta_struct','grad_phi_struct','theta_struct','phi_struct','e');
        
    elseif find_orientation && ~find_coefficients
        save([filepath 'current_opt_saxs_angle.mat'], ...
            'grad_theta_struct','grad_phi_struct','theta_struct','phi_struct','e');
        
    elseif ~find_orientation && find_coefficients
        if order == 1
            save([filepath 'current_opt_symsaxs.mat'], 'a_tomo', 'grad_a', 'e');
        else
            save([filepath 'current_opt_saxs.mat'], 'a_tomo', 'grad_a', 'e');
        end
    end
end

