%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [proj_out,E,Reg,Ereg] = plot_projection_reg(params, projection)

%*-------------------------------------------------------------------------------------*
%|                                                                                                           |
%|  Except where otherwise noted, this work is licensed under a            |
%|  Creative Commons Attribution-NonCommercial-ShareAlike 4.0        |
%|  International (CC BY-NC-SA 4.0) license.                                         |
%|                                                                                                           |
%|  Copyright (c) 2017 by Paul Scherrer Institute (http://www.psi.ch)       |
%|                                                                                                           |
%|      Author: CXS group, PSI                                                                |
%*------------------------------------------------------------------------------------*
% You may use this code with the following provisions:
%
% If this code, or subfunctions or parts of it, is used for research in a 
%   publication or if it is fully or partially rewritten for another 
%   computing language the authors and institution should be acknowledged 
%   in written form and additionally you should cite:
%     M. Liebi, M. Georgiadis, A. Menzel, P. Schneider, J. Kohlbrecher, 
%     O. Bunk, and M. Guizar-Sicairos, “Nanostructure surveys of 
%     macroscopic specimens by small-angle scattering tensor tomography,”
%     Nature 527, 349-352 (2015).   (doi:10.1038/nature16056)
%
% A publication that focuses on describing features, or parameters, that
%    are already existing in the code should be first discussed with the
%    authors.
%   
% This code and subroutines are part of a continuous development, they 
%    are provided “as they are” without guarantees or liability on part
%    of PSI or the authors. It is the user responsibility to ensure its 
%    proper use and the correctness of the results.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%% parameters for the optimization
phi_det = params.phi_det;                           % Measurement angles (detector)
l = params.l;                                       % Polar order of the spherical harmonics
order = numel(l);
lines = params.lines;                            % dimension of scanned area
points_per_line = params.points_per_line;        % dimension of scanned area
numOfsegments = params.numsegments;                     % number of segements the saxs data is averaged (default is 16)
numOfvoxels = params.numOfvoxels;           % dimension of tomogram
numOfpixels = length(projection)*points_per_line*lines;

%parameters for arb_projection
p.volume_upsampling = params.volume_upsampling;     % Upsamples the volume by a factor of 2 to reduce artifacts (recommended setting =1)
p.method = params.method;                           % 'nearest' or 'bilinear' method for volume allocation
p.filter_2D = params.filter_2D;                     % Strength of filter applied to image (recommended setting =1)

%opt_inputs is the optimization vector, containing theta_struct, phi_struct and a, the coefficents of the spherical harmonics
theta_struct = params.theta_0;
phi_struct = params.phi_0;
a = reshape(params.a, numOfvoxels, order);
fprintf('start loop over projections in iteration %d \n', length(errorplot([])));

% Precalculate parameters
arr_phi_det = [cos(phi_det); sin(phi_det)];

sin_theta_struct = reshape(sin(theta_struct), 1, 1, numOfvoxels);
cos_theta_struct = reshape(cos(theta_struct), 1, 1, numOfvoxels);
sin_phi_struct = reshape(sin(phi_struct), 1, 1, numOfvoxels);
cos_phi_struct = reshape(cos(phi_struct), 1, 1, numOfvoxels);

zeros_struct = zeros(1, 1, numOfvoxels);

%%% Define the rotation matrices
% Rot_str is the combination of 1. Rotation around y with theta_struct
% and 2. Rotation around z with theta_struct
% this rotation is needed from the spherical harmonics coordinate system (main
% orientation axis along z-axis) to the object coordinate system
Rot_str = [ ...
    cos_theta_struct.*cos_phi_struct, cos_theta_struct.*sin_phi_struct, -sin_theta_struct; ...
    -sin_phi_struct                 , cos_phi_struct                  , zeros_struct     ; ...
    sin_theta_struct.*cos_phi_struct, sin_theta_struct.*sin_phi_struct, cos_theta_struct];


% Determine Ylm and Ylm+1 coefficients
Ylm_coef = ...
    [sqrt(1/pi)*[  1/2,      0,       0,      0]; ...
    sqrt(5/pi) *[ -1/4,    3/4,       0,      0]; ...
    sqrt(1/pi) *[ 9/16,  -45/8,  105/16,      0]; ...
    sqrt(13/pi)*[-5/32, 105/32, -315/32, 231/32]];
Ylm_coef = Ylm_coef(1:order, 1:order);

a_temp = reshape(a.', 1, order, []);

YlmPLUS1_coef = ...
    [                [0,     0,     0,      0]; ...
    sqrt(15/(2*pi)) *[0,  -1/2,     0,      0]; ...
    sqrt(5/pi)      *[0,   9/8, -21/8,      0]; ...
    sqrt(273/(2*pi))*[0, -5/16,  15/8, -33/16]];
YlmPLUS1_coef = YlmPLUS1_coef(1:order, 1:order);

a_temp2 = bsxfun(@times, a, sqrt(l.*(l+1))) * YlmPLUS1_coef;
a_temp2 = reshape(a_temp2.', 1, order, []);

ones_struct = ones(1, numOfsegments, numOfvoxels);

%%%%changed
if params.regularization_angle
    theta_tomo=reshape(theta_struct,points_per_line,lines,lines);
    phi_tomo=reshape(phi_struct,points_per_line,lines,lines);
    sin_theta_tomo=reshape(sin_theta_struct,points_per_line,lines,lines);
    cos_theta_tomo=reshape(cos_theta_struct,points_per_line,lines,lines);
    sin_phi_tomo=reshape(sin_phi_struct,points_per_line,lines,lines);
    cos_phi_tomo=reshape(cos_phi_struct,points_per_line,lines,lines);
    dot_ind1=ones(points_per_line,lines,lines);
    dot_ind2=ones(points_per_line,lines,lines);
    dot_ind3=ones(points_per_line,lines,lines);
    one_matrix=ones(points_per_line,lines,lines);
    dot_ind1(1:end-1,:,:)=(sin_theta_tomo(1:end-1,:,:).*sin_theta_tomo(2:end,:,:).*cos(phi_tomo(1:end-1,:,:)-phi_tomo(2:end,:,:))+cos_theta_tomo(1:end-1,:,:).*cos_theta_tomo(2:end,:,:));
    dot_ind2(:,1:end-1,:)=(sin_theta_tomo(:,1:end-1,:).*sin_theta_tomo(:,2:end,:).*cos(phi_tomo(:,1:end-1,:)-phi_tomo(:,2:end,:))+cos_theta_tomo(:,1:end-1,:).*cos_theta_tomo(:,2:end,:));
    dot_ind3(:,:,1:end-1)=(sin_theta_tomo(:,:,1:end-1).*sin_theta_tomo(:,:,2:end).*cos(phi_tomo(:,:,1:end-1)-phi_tomo(:,:,2:end))+cos_theta_tomo(:,:,1:end-1).*cos_theta_tomo(:,:,2:end));
    Reg=sum(sum(sum(one_matrix-abs(dot_ind1)+one_matrix-abs(dot_ind2)+one_matrix-abs(dot_ind3))));
    
    %%free memory
    theta_tomo=[];
    phi_tomo=[];
    dot_ind1=[];
    dot_ind2=[];
    dot_ind3=[];
    one_matrix=[];
    sin_theta_tomo=[];
    cos_theta_tomo=[];
    sin_phi_tomo=[];
    cos_phi_tomo=[];
    
end
%%%%changed
E=0;
parfor ii = 1:length(projection) %use parallel processing for the loop over all projections
    data = double(projection(ii).data);
    Rot_exp_now = double(projection(ii).Rot_exp);
    Rot_exp_now_mod = Rot_exp_now(1:2, :)' * arr_phi_det; %transpose to be able to calculate
    sh_cut = bsxpagemult(Rot_str, Rot_exp_now_mod);
    cos_theta_sh_cut = sh_cut(3, :, :);
    block = [ones_struct; cumprod(repmat(cos_theta_sh_cut.^2, order-1, 1), 1)];
    Ylm = bsxpagemult(Ylm_coef, block);
    Ymn_2D = bsxpagemult(a_temp, Ylm);
    data_synt_vol = permute(abs(Ymn_2D.^2), [3, 2, 1]); % modeled intensities
    Ymn_aux_vol = permute(real(repmat(conj(Ymn_2D), order, 1, 1) .* Ylm), [3, 2, 1]);
    Ylm = []; %#ok<*NASGU> % free up memory
    sin_theta_sh_cut = sqrt(1-cos_theta_sh_cut.^2);
    phi_sh_cut = atan2(sh_cut(2, :, :), sh_cut(1, :, :));
    sh_cut = []; % free up memory
    const_val = exp(1i.*phi_sh_cut) .* sin_theta_sh_cut .* cos_theta_sh_cut;
    cos_theta_sh_cut = []; % free up memory
    block2 = [ones_struct; repmat(const_val, order-1, 1, 1) .* block(1:order-1, :, :)];
    block = []; % free up memory
    const_val = [];
    YlmPLUS1 = bsxpagemult(a_temp2, block2);
    block2 = []; % free up memory
    YlmPLUS1 = []; % free up memory
    Ymn_2D = [];
    phi_sh_cut = [];
    sin_theta_sh_cut = [];
       
    %%%%reshape all the data arrays into dimensions of the array, from
    %numOfvoxels to (lines x lines x points_per_line)
    data_synt_vol = reshape(data_synt_vol, points_per_line,lines,lines, numOfsegments);
    Ymn_aux_vol = reshape(Ymn_aux_vol, points_per_line,lines,lines, numOfsegments, order);
    
    %%%%projection of the data_synt_vol onto detector plane, in direction of
    %Rot_exp
    N = [points_per_line,lines,lines]; % Size of tomogram
    x = (1:N(2)) - ceil(N(2)/2);
    y = (1:N(1)) - ceil(N(1)/2);
    z = (1:N(3)) - ceil(N(3)/2);
    [X, Y, Z] = meshgrid(x, y, z); % (x,y,z) (2,1,3)
    % If x is defined as the 3rd dimension N(3), then it should appear in the
    % third position of the meshgrid command. If x is defined as the first
    % dimension N(1), then it appears in the second position of the meshgrid,
    % because meshgrid is (x,y,z) and size is (y,x,z)
    
    % output of projection should have the size of the measured projection
    %put in alignment correction values from registration
    xout = (1:size(data,2)) - ceil(size(data,2)/2);
    yout = (1:size(data,1)) - ceil(size(data,1)/2);
    
    [proj_out_all, xout, yout] = arb_projection(data_synt_vol, X, Y, Z, Rot_exp_now, p, xout, yout);
    
    data_synt_vol = []; % free up memory
    
    %%%caluculating the error (difference between estimated intensity and measured data), taking poisson noise into account
    %%%correct with diode diode data
    %%%include window function
    %auxillary functions, used in E and the gradient
    % here include window function (used in gradient and errormetric) and correct for absorption with diode data
    aux_diff_poisson = (sqrt(proj_out_all) - sqrt(data)).* projection(ii).window_mask;
    error_norm = (2*sum(sum(sum(aux_diff_poisson.^2))))./numOfpixels;
    E = E + error_norm; % error metric with poisson noise
    proj_out(ii).projection=proj_out_all;
    proj_out(ii).rotx = projection(ii).rot_x;
    proj_out(ii).roty = projection(ii).rot_y;
    proj_out(ii).errorplot = aux_diff_poisson;
    proj_out(ii).error = error_norm; % error metric with poisson noise
end

%%%changed
if params.regularization_angle
    Ereg = params.regularization_angle_coeff.*Reg./numOfvoxels;
end
